#include "roomlistmodel.h"

#include <algorithm>

#include <QDebug>

QModelIndex RoomListModel::calcIndex(std::string id) {
    int pos = 0;
    for (const auto &r : rooms) {
        if (r->id == id)
            return index(pos, 0);
        else
            pos++;
    }

    return index(-1, 0);
}

void RoomListModel::addRoom(QSharedPointer<Room> room) {
    int posIdx = (int)(rooms.size());
    beginInsertRows(QModelIndex(), posIdx, posIdx);
    rooms.push_back(room);
    endInsertRows();

    roomById[room->id] = room;
}

void RoomListModel::updateRoom(QSharedPointer<Room> room) {
    roomById[room->id] = room;

    auto idx = calcIndex(room->id);
    if (idx.row() != -1)
        emit dataChanged(idx, idx);
}

QHash<int, QByteArray> RoomListModel::roleNames() const {
    QHash<int, QByteArray> roles;
    roles[RoomModelRole] = "roomModel";
    roles[RoomNameRole] = "roomName";
    roles[UnreadNotificationsRole] = "unreadNotifications";
    roles[LastMessageRole] = "lastMessage";
    roles[RoomPictureRole] = "roomPicture";
    roles[TagRole] = "tag";
    roles[IsInvite] = "isInvite";
    return roles;
}

int RoomListModel::rowCount(const QModelIndex &parent) const {
    Q_UNUSED(parent);
    return (int)rooms.size();
}

QVariant RoomListModel::data(const QModelIndex &index, int role) const {

    if (index.row() < 0 || index.row() >= rowCount())
        return QVariant();

    const auto &room = [this, index]() { return rooms[index.row()]; }();

    switch (role) {
    case RoomModelRole:
        return QVariant::fromValue(room.data());
    case RoomNameRole:
        return QString::fromStdString(room->name());
    case UnreadNotificationsRole:
        return room->unreadNotifications;
    case LastMessageRole:
        return room->msgInfo.body;
    case RoomPictureRole:
        return QString::fromStdString(room->avatar_url).replace("mxc://", "image://MXCImage/");
    case TagRole:
        return QString::fromStdString(room->tags.empty() ? "" : room->tags.front());
    case IsInvite:
        return room->is_invite;

    default:
        return QVariant();
    }
}

RoomListModel::RoomListModel(QObject *parent) { Q_UNUSED(parent); }

bool RoomListSortProxy::lessThan(const QModelIndex &rLeft, const QModelIndex &rRight) const {
    QModelIndex const left_idx = sourceModel()->index(rLeft.row(), 0, QModelIndex());
    QModelIndex const right_idx = sourceModel()->index(rRight.row(), 0, QModelIndex());

    auto left = [this, left_idx](RoomListModel::RoomRoles role) -> QVariant {
        return sourceModel()->data(left_idx, role);
    };
    auto right = [this, right_idx](RoomListModel::RoomRoles role) -> QVariant {
        return sourceModel()->data(right_idx, role);
    };

    if (left(RoomListModel::IsInvite) != right(RoomListModel::IsInvite))
        return left(RoomListModel::IsInvite) > right(RoomListModel::IsInvite);
    else {
        if (left(RoomListModel::TagRole).toString() != right(RoomListModel::TagRole).toString()) {
            return QString::localeAwareCompare(left(RoomListModel::TagRole).toString(),
                                               right(RoomListModel::TagRole).toString()) < 0;
        } else {
            if (left(RoomListModel::UnreadNotificationsRole) != right(RoomListModel::UnreadNotificationsRole))
                return left(RoomListModel::UnreadNotificationsRole) > right(RoomListModel::UnreadNotificationsRole);
            else {
                auto leftTime = left(RoomListModel::RoomModelRole).value<Room *>()->msgInfo.datetime;
                auto rightTime = right(RoomListModel::RoomModelRole).value<Room *>()->msgInfo.datetime;
                if (leftTime != rightTime) {
                    return leftTime > rightTime;
                } else
                    return rLeft.row() < rRight.row();
            }
        }
    }
}
