#include "sync.h"

#include <chrono>
#include <thread>

#include <contextproperty.h>

#include <QtDebug>

#include "client.h"
#include "debug_out.h"

Sync::Sync(RoomListModel &rooms, Login *login, QObject *parent)
    : QThread(parent), rooms(rooms), stop(false), exit(false) {
    static ContextProperty *prop = new ContextProperty("Internet.NetworkState");

    connect(this, &Sync::nextSync, this, &Sync::sync, Qt::QueuedConnection);
    connect(prop, &ContextProperty::valueChanged, [this]() {
        qDebug() << "Value changed: " << prop->value();

        if (prop->value() == QString("connected")) {
            if (this->stop)
                emit startSync();
        } else {
            emit stopSync();
        }
    });
}

void Sync::startSync() {
    qDebug() << "Start sync";
    this->stop = false;
    if (!isRunning()) {
        start(LowPriority);
    } else {
        condition.wakeOne();
    }
}

Sync::~Sync() {}

void Sync::exitSync() {
    qDebug() << "Exit sync";
    this->exit = true;
    this->stop = true;
    http::client().shutdown();
    condition.wakeOne();
    quit();
}

void Sync::stopSync() {
    qDebug() << "Stop sync";
    // this->exit = true;
    this->stop = true;
    http::client().shutdown();
    condition.wakeOne();
    quit();
    wait();
}

void Sync::run() { emit nextSync(); }

template <class T> auto strippedToStateEv(const mtx::events::StrippedEvent<T> &stripped) -> mtx::events::StateEvent<T> {
    mtx::events::StateEvent<T> ev;
    ev.state_key = stripped.state_key;
    ev.sender = stripped.sender;
    ev.content = stripped.content;
    ev.type = stripped.type;
    return ev;
}

void Sync::sync() {
    qDebug() << "sync";

    if (this->stop) {
        if (this->exit) {
            qDebug() << "sync exiting";
            return;
        } else {
            qDebug() << "sync pausing";
            mutex.lock();
            this->condition.wait(&mutex);
            mutex.unlock();
        }
    }

    mtx::http::SyncOpts opts = {};
    opts.since = since;
    if (since.empty())
        opts.full_state = true;

    http::client().sync(opts, [this](const mtx::responses::Sync &res, mtx::http::RequestErr err) {
        if (err) {
            qDebug() << "Failed sync";
            qDebug() << *err;

            if (err->error_code == boost::system::errc::network_unreachable) {
                qDebug() << "Network changed, trying again";
                emit nextSync();
            }

            return;
        }

        this->since = res.next_batch;

        for (const std::pair<std::string, mtx::responses::JoinedRoom> &join : res.rooms.join) {
            const mtx::responses::JoinedRoom &r = join.second;

            bool createRoom = false;
            QSharedPointer<Room> room = rooms.room(join.first);
            if (!room) {
                room = QSharedPointer<Room>(new Room);
                room->id = join.first;
                room->is_invite = false;
                createRoom = true;
            }

            qDebug() << QString::fromStdString(room->id);

            for (const auto& e : r.account_data.events)
            {
                if (const auto t = boost::get<mtx::events::Event<mtx::events::account_data::Tag>>(&e)) {
                    qDebug() << "Tag event";
                    std::vector<std::string> tags;
                    for (const auto &tag : t->content.tags)
                        tags.push_back(tag.first);
                    room->applyTags(tags);
                }
            }

            std::vector<mtx::events::collections::TimelineEvents> timeline;

            for (const auto &e : r.state.events) {
                boost::apply_visitor(
                    [&timeline](const auto &ev) { timeline.push_back(mtx::events::collections::TimelineEvents{ev}); },
                    e);
            }
            for (const auto &e : r.timeline.events) {
                boost::apply_visitor(
                    [&timeline](const auto &ev) { timeline.push_back(mtx::events::collections::TimelineEvents{ev}); },
                    e);
            }

            room->applyEvents(timeline, r.timeline.prev_batch, res.next_batch);
            room->applyUnreadNotifications(r.unread_notifications.notification_count);

            if (createRoom)
                emit newRoom(room);
            else
                emit roomUpdated(room);
        }

        for (const std::pair<std::string, mtx::responses::InvitedRoom> &invite : res.rooms.invite) {
            const mtx::responses::InvitedRoom &r = invite.second;

            bool createRoom = false;
            QSharedPointer<Room> room = rooms.room(invite.first);
            if (!room) {
                room = QSharedPointer<Room>(new Room);
                createRoom = true;

                room->id = invite.first;
                room->tags.push_back("invite");
                room->is_invite = true;
                room->name_ = r.name();
                room->avatar_url = r.avatar();
            }

            qDebug() << QString::fromStdString(room->id);

            std::vector<mtx::events::collections::TimelineEvents> timeline;
            for (const auto &e : r.invite_state) {
                boost::apply_visitor(
                    [&timeline](const auto &ev) {
                        mtx::events::collections::TimelineEvents tev = strippedToStateEv(ev);
                        timeline.push_back(tev);
                    },
                    e);
            }

            room->applyEvents(timeline, "", res.next_batch);

            if (createRoom)
                emit newRoom(room);
            else
                emit roomUpdated(room);
        }

        emit nextSync();
    });
}
